import unittest

try:
    import matplotlib
except ImportError:
    matplotlib = None

from openmdao.utils.testing_utils import use_tempdirs


@use_tempdirs
class TestBrachistochronePolynomialControl(unittest.TestCase):

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_gauss_lobatto(self):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True, solve_segments=False)

        phase.add_state('y', fix_initial=True, fix_final=True, solve_segments=False)

        phase.add_state('v', fix_initial=True, fix_final=False, solve_segments=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_radau(self):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Radau(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_birkhoff(self):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Birkhoff(num_nodes=15))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()


@use_tempdirs
class TestBrachistochronePolynomialControlBoundaryConstrained(unittest.TestCase):

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_gauss_lobatto(self):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_boundary_constraint('theta', loc='initial', lower=0, upper=1.0)
        phase.add_boundary_constraint('theta', loc='final', lower=100, upper=105.)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_radau(self):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Radau(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_boundary_constraint('theta', loc='initial', lower=0, upper=1.0)
        phase.add_boundary_constraint('theta', loc='final', lower=100, upper=105.)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_birkhoff(self):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Birkhoff(num_nodes=15))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_boundary_constraint('theta', loc='initial', lower=0, upper=1.0)
        phase.add_boundary_constraint('theta', loc='final', upper=105.0)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()


@use_tempdirs
class TestBrachistochronePolynomialControlPathConstrained(unittest.TestCase):

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_gauss_lobatto(self):
        import matplotlib.pyplot as plt
        plt.switch_backend('Agg')
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('theta', lower=0, upper=120)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_radau(self):
        import matplotlib.pyplot as plt
        plt.switch_backend('Agg')
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Radau(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('theta', lower=1, upper=120)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_birkhoff(self):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Birkhoff(num_nodes=15))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('theta', lower=1, upper=120)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()


@use_tempdirs
class TestBrachistochronePolynomialControlRatePathConstrained(unittest.TestCase):

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_gauss_lobatto(self):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('theta_rate', lower=0, upper=120)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_radau(self):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Radau(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('theta_rate', lower=0, upper=120)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_birkhoff(self):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Birkhoff(num_nodes=15))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('theta_rate', lower=0, upper=120)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()


@use_tempdirs
class TestBrachistochronePolynomialControlRate2PathConstrained(unittest.TestCase):

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_gauss_lobatto(self):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('theta_rate2', lower=-0.01, upper=0.01)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_radau(self):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Radau(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('theta_rate2', lower=-0.01, upper=0.01)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_birkhoff(self):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Birkhoff(num_nodes=15))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('theta_rate2', lower=-0.01, upper=0.01)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        fig, ax = plt.subplots()
        fig.suptitle('Brachistochrone Solution')

        x_imp = p.get_val('phase0.timeseries.x')
        y_imp = p.get_val('phase0.timeseries.y')

        x_exp = exp_out.get_val('phase0.timeseries.x')
        y_exp = exp_out.get_val('phase0.timeseries.y')

        ax.plot(x_imp, y_imp, 'ro', label='solution')
        ax.plot(x_exp, y_exp, 'b-', label='simulated')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.grid(True)
        ax.legend(loc='upper right')

        fig, ax = plt.subplots()

        t_imp = p.get_val('phase0.timeseries.time')
        theta_imp = p.get_val('phase0.timeseries.theta')

        ax.plot(t_imp, theta_imp, 'ro', label='solution')

        ax.set_xlabel('time (s)')
        ax.set_ylabel(r'$\theta$ (deg)')
        ax.grid(True)
        ax.legend(loc='upper right')

        plt.show()


@use_tempdirs
class TestBrachistochronePolynomialControlSimulation(unittest.TestCase):

    def test_brachistochrone_polynomial_control_gauss_lobatto(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        theta_imp = p.get_val('phase0.timeseries.theta')
        theta_exp = exp_out.get_val('phase0.timeseries.theta')

        assert_near_equal(theta_exp[0], theta_imp[0])
        assert_near_equal(theta_exp[-1], theta_imp[-1])

    def test_brachistochrone_polynomial_control_radau(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Radau(num_segments=10))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        theta_imp = p.get_val('phase0.timeseries.theta')
        theta_exp = exp_out.get_val('phase0.timeseries.theta')

        assert_near_equal(theta_exp[0], theta_imp[0])
        assert_near_equal(theta_exp[-1], theta_imp[-1])

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_polynomial_control_birkhoff(self):
        import matplotlib
        matplotlib.use('Agg')
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Birkhoff(num_nodes=15))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])

        # Solve for the optimal trajectory
        p.run_driver()

        # Test the results
        assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)

        # Generate the explicitly simulated trajectory
        exp_out = phase.simulate()

        theta_imp = p.get_val('phase0.timeseries.theta')
        theta_exp = exp_out.get_val('phase0.timeseries.theta')

        assert_near_equal(theta_exp[0], theta_imp[0])
        assert_near_equal(theta_exp[-1], theta_imp[-1])


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np

import openmdao.api as om

from ..transcription_base import TranscriptionBase
from ..common import TimeComp, TimeseriesOutputComp
from .components import BirkhoffIterGroup, BirkhoffBoundaryGroup

from ..grid_data import BirkhoffGrid
from dymos.utils.misc import get_rate_units, _format_phase_constraint_alias
from dymos.utils.introspection import get_promoted_vars, get_source_metadata
from dymos.utils.indexing import get_constraint_flat_idxs, get_src_indices_by_row


class Birkhoff(TranscriptionBase):
    """
        Birkhoff Pseudospectral Transcription.

        Parameters
        ----------
        **kwargs : dict
            Dictionary of optional arguments.

        References
        ----------
        I. M. Ross, "A Universeal Birkhoff Theory for Fast Trajectory Optimization"
        https://arxiv.org/abs/2308.01400v2
        """
    def __init__(self, **kwargs):
        super(Birkhoff, self).__init__(**kwargs)
        self._rhs_source = 'ode_iter_group.ode_all'

    def initialize(self):
        """
        Declare transcription options.
        """
        self.options.declare('num_nodes', types=int, default=3,
                             desc='The number of nodes in the grid')

        self.options.declare('grid_type', values=('cgl', 'lgl'), default='cgl',
                             desc='Specifies which type of grid is to be used. '
                                  'Options are Chebyshev-Gauss-Lobatto ("cgl") '
                                  'and Legendre-Gauss-Lobatto ("lgl")')

        self.options.declare(name='solve_segments', default=False,
                             values=(False, 'forward', 'backward'),
                             desc='Applies \'solve_segments\' behavior to _all_ states in the Phase. '
                                  'If \'forward\', collocation defects within each '
                                  'segment are solved with a Newton solver by fixing the initial value in the '
                                  'phase (if using compressed transcription) or segment (if not using '
                                  'compressed transcription). This provides a forward shooting (or multiple shooting) '
                                  'method.  If \'backward\', the final value in the phase or segment is fixed '
                                  'and a solver finds the other ones to mimic reverse propagation. Set '
                                  'to False (the default) to explicitly disable the use of a solver to '
                                  'converge the state time history.')

    def init_grid(self):
        """
        Setup the GridData object for the Transcription.
        """
        self.grid_data = BirkhoffGrid(num_nodes=self.options['num_nodes'],
                                      grid_type=self.options['grid_type'])

    def setup_time(self, phase):
        """
        Setup the time component.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        grid_data = self.grid_data

        super(Birkhoff, self).setup_time(phase)

        time_comp = TimeComp(num_nodes=grid_data.num_nodes, node_ptau=grid_data.node_ptau,
                             node_dptau_dstau=grid_data.node_dptau_dstau,
                             units=phase.time_options['units'],
                             initial_val=phase.time_options['initial_val'],
                             duration_val=phase.time_options['duration_val'])

        phase.add_subsystem('time', time_comp, promotes_inputs=['*'], promotes_outputs=['*'])

    def configure_time(self, phase):
        """
        Configure the inputs/outputs on the time component.

        This method assumes that target introspection has already been performed by the phase and thus
        options['targets'], options['time_phase_targets'], options['t_initial_targets'],
        and options['t_duration_targets'] are all correctly populated.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(Birkhoff, self).configure_time(phase)
        phase.time.configure_io()
        options = phase.time_options
        ode = phase._get_subsystem(self._rhs_source)
        ode_inputs = get_promoted_vars(ode, iotypes='input')

        # The tuples here are (name, user_specified_targets, dynamic)
        for name, targets in [('t', options['targets']),
                              ('t_phase', options['time_phase_targets'])]:
            if targets:
                src_idxs = self.grid_data.subset_node_indices['all']
                phase.connect(name, [f'ode_all.{t}' for t in targets], src_indices=src_idxs,
                              flat_src_indices=True)

        for name, targets in [('t_initial', options['t_initial_targets']),
                              ('t_duration', options['t_duration_targets']),
                              ('t_final', options['t_final_targets'])]:
            for t in targets:
                shape = ode_inputs[t]['shape']

                if shape == (1,):
                    src_idxs = endpoint_src_idxs = None
                    flat_src_idxs = None
                    src_shape = None
                else:
                    src_idxs = np.zeros(self.grid_data.subset_num_nodes['all'], dtype=int)
                    endpoint_src_idxs = np.zeros(2, dtype=int)
                    flat_src_idxs = True
                    src_shape = (1,)

                phase.promotes('ode_all', inputs=[(t, name)], src_indices=src_idxs,
                               flat_src_indices=flat_src_idxs, src_shape=src_shape)

                phase.promotes('boundary_vals', inputs=[(t, name)], src_indices=endpoint_src_idxs,
                               flat_src_indices=flat_src_idxs, src_shape=src_shape)
            if targets:
                phase.set_input_defaults(name=name,
                                         val=np.ones((1,)),
                                         units=options['units'])

    def setup_states(self, phase):
        """
        Setup the states for this transcription.

        In the Birkhoff transcription, everything typically done in this
        method is instead done by the BirkhoffIterGroup.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        self.any_solved_segs = False
        self.any_connected_opt_segs = False
        for options in phase.state_options.values():
            # Transcription solve_segments overrides state solve_segments if its not set
            if options['solve_segments'] is None:
                options['solve_segments'] = self.options['solve_segments']

            if options['solve_segments']:
                self.any_solved_segs = True
            elif options['input_initial']:
                self.any_connected_opt_segs = True

    def configure_controls(self, phase):
        """
        Configure the inputs/outputs for the controls.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super().configure_controls(phase)

        if phase.control_options:
            phase.control_comp.configure_io()
            phase.promotes('control_comp',
                           any=['*controls:*', '*control_values:*', '*control_rates:*'])

            phase.connect('dt_dstau', 'control_comp.dt_dstau')

        for name, options in phase.control_options.items():
            if options['targets']:
                phase.connect(f'control_values:{name}', [f'ode_all.{t}' for t in options['targets']])
                phase.connect(f'control_values:{name}', [f'boundary_vals.{t}' for t in options['targets']],
                              src_indices=om.slicer[[0, -1], ...])

            if options['rate_targets']:
                phase.connect(f'control_rates:{name}_rate',
                              [f'ode_all.{t}' for t in options['rate_targets']])
                phase.connect(f'control_rates:{name}_rate',
                              [f'boundary_vals.{t}' for t in options['rate_targets']],
                              src_indices=om.slicer[[0, -1], ...])

            if options['rate2_targets']:
                phase.connect(f'control_rates:{name}_rate2',
                              [f'ode_all.{t}' for t in options['rate2_targets']])
                phase.connect(f'control_rates:{name}_rate2',
                              [f'boundary_vals.{t}' for t in options['rate2_targets']],
                              src_indices=om.slicer[[0, -1], ...])

    def setup_ode(self, phase):
        """
        Setup the ode for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """

        ODEClass = phase.options['ode_class']
        grid_data = self.grid_data

        ode_init_kwargs = phase.options['ode_init_kwargs']

        phase.add_subsystem('ode_iter_group',
                            subsys=BirkhoffIterGroup(grid_data=grid_data, state_options=phase.state_options,
                                                     time_options=phase.time_options,
                                                     parameter_options=phase.parameter_options,
                                                     ode_class=ODEClass,
                                                     ode_init_kwargs=ode_init_kwargs,
                                                     calc_exprs=phase._calc_exprs),
                            promotes=['*'])

        phase.add_subsystem('boundary_vals',
                            subsys=BirkhoffBoundaryGroup(grid_data=grid_data,
                                                         ode_class=ODEClass,
                                                         ode_init_kwargs=ode_init_kwargs,
                                                         calc_exprs=phase._calc_exprs,
                                                         parameter_options=phase.parameter_options),
                            promotes_inputs=['initial_states:*', 'final_states:*'])

    def configure_ode(self, phase):
        """
        Create connections to the introspected states.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        phase._get_subsystem('boundary_vals').configure_io(phase)
        phase._get_subsystem('ode_iter_group').configure_io(phase)

    def setup_defects(self, phase):
        """
        Create the continuity_comp to house the defects.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_defects(self, phase):
        """
        Connect the collocation constraints.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        num_nodes = self.grid_data.subset_num_nodes['all']
        for name, options in phase.state_options.items():
            rate_source_type = phase.classify_var(options['rate_source'])
            rate_src_path = self._get_rate_source_path(name, phase)
            if rate_src_path.startswith('parameter_vals:'):
                src_idxs = om.slicer[np.zeros(num_nodes, dtype=int), ...]
            else:
                src_idxs = None

            if rate_source_type not in ('state', 'ode'):
                phase.connect(rate_src_path, f'f_computed:{name}', src_indices=src_idxs)

    def setup_solvers(self, phase):
        """
        Setup the solvers.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_solvers(self, phase, requires_solvers=None):
        """
        Configure the solvers.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        requires_solvers : dict[str: bool]
            A dictionary mapping a string descriptor of a reason why a solver is required,
            and whether a solver is required.
        """
        ode_iter_group = phase._get_subsystem('ode_iter_group')
        req_solvers = {'implicit outputs': ode_iter_group._implicit_outputs}

        if requires_solvers is not None:
            req_solvers.update(requires_solvers)

        super().configure_solvers(phase, requires_solvers=req_solvers)

    def configure_timeseries_outputs(self, phase):
        """
        Create connections from time series to all post-introspection sources.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        for timeseries_name, timeseries_options in phase._timeseries.items():
            timeseries_comp = phase._get_subsystem(timeseries_name)
            ts_inputs_to_promote = []
            for input_name, src, src_idxs in timeseries_comp._configure_io(timeseries_options):
                # If the src was added, promote it if it was a state,
                # or connect it otherwise.
                if src.startswith('states:'):
                    state_name = src.split(':')[-1]
                    ts_inputs_to_promote.append((input_name, f'states:{state_name}'))
                else:
                    phase.connect(src_name=src,
                                  tgt_name=f'{timeseries_name}.{input_name}',
                                  src_indices=src_idxs)
            phase.promotes(timeseries_name, inputs=ts_inputs_to_promote)

    def setup_timeseries_outputs(self, phase):
        """
        Setup the timeseries for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        gd = self.grid_data

        for name, options in phase._timeseries.items():
            if options['transcription'] is None:
                ogd = None
            else:
                ogd = options['transcription'].grid_data

            timeseries_comp = TimeseriesOutputComp(input_grid_data=gd,
                                                   output_grid_data=ogd,
                                                   output_subset=options['subset'],
                                                   time_units=phase.time_options['units'])
            phase.add_subsystem(name, subsys=timeseries_comp)

            phase.connect('dt_dstau', f'{name}.dt_dstau', flat_src_indices=True)

    def _get_constraint_kwargs(self, constraint_type, options, phase):
        """
        Given the constraint options provide the keyword arguments for the OpenMDAO add_constraint method.

        Parameters
        ----------
        constraint_type : str
            One of 'initial', 'final', or 'path'.
        options : dict
            The constraint options.
        phase : Phase
            The dymos phase to which the constraint applies.

        Returns
        -------
        con_output : str
            The phase-relative path being constrained.
        constraint_kwargs : dict
            Keyword arguments for the OpenMDAO add_constraint method.
        """
        num_nodes = self._get_num_timeseries_nodes()

        constraint_kwargs = {key: options for key, options in options.items()}

        # Determine the path to the variable which we will be constraining
        var = options['name']
        var_type = phase.classify_var(var)

        # These are the flat indices at a single point in time used
        # in either initial, final, or path constraints.
        idxs_in_initial = phase._indices_in_constraints(var, 'initial')
        idxs_in_final = phase._indices_in_constraints(var, 'final')
        idxs_in_path = phase._indices_in_constraints(var, 'path')

        size = np.prod(options['shape'], dtype=int)

        flat_idxs = get_constraint_flat_idxs(options)

        # Now we need to convert the indices given by the user at any given point
        # to flat indices to be given to OpenMDAO as flat indices spanning the phase.
        if var_type == 'parameter':
            if any([idxs_in_initial.intersection(idxs_in_final),
                    idxs_in_initial.intersection(idxs_in_path),
                    idxs_in_final.intersection(idxs_in_path)]):
                raise RuntimeError(f'In phase {phase.pathname}, parameter `{var}` is subject to multiple boundary '
                                   f'or path constraints.\nParameters are single values that do not change in '
                                   f'time, and may only be used in a single boundary or path constraint.')
            constraint_kwargs['indices'] = flat_idxs
        else:
            if constraint_type == 'initial':
                constraint_kwargs['indices'] = flat_idxs
            elif constraint_type == 'final':
                constraint_kwargs['indices'] = size + flat_idxs
            else:
                # Path
                path_idxs = []
                for i in range(num_nodes):
                    path_idxs.extend(size * i + flat_idxs)

                constraint_kwargs['indices'] = path_idxs

        con_path = constraint_kwargs.pop('constraint_path')
        con_name = constraint_kwargs.pop('constraint_name')

        constraint_kwargs['alias'] = _format_phase_constraint_alias(phase, con_name,
                                                                    constraint_type,
                                                                    options['indices'])
        constraint_kwargs.pop('name')
        constraint_kwargs.pop('shape')
        constraint_kwargs['flat_indices'] = True

        return con_path, constraint_kwargs

    def _get_response_src(self, var, loc, phase, ode_outputs=None):
        """
        Return the path to the variable that will be used as a response..

        Parameters
        ----------
        var : str
            Name of the variable to be used as the response.
        loc : str
            The location of the response in the phase ['initial', 'final'].
        phase : dymos.Phase
            Phase object containing in which the objective resides.
        ode_outputs : dict or None
            A dictionary of ODE outputs as returned by get_promoted_vars.

        Returns
        -------
        obj_path : str
            Path to the source.
        shape : tuple
            Source shape.
        units : str
            Source units.
        linear : bool
            True if the objective quantity is linear.
        """
        time_units = phase.time_options['units']
        var_type = phase.classify_var(var)

        if ode_outputs is None:
            ode_outputs = get_promoted_vars(phase._get_subsystem(self._rhs_source), 'output')

        if var_type == 't':
            shape = (1,)
            units = time_units
            linear = True
            constraint_path = 't'
        elif var_type == 't_phase':
            shape = (1,)
            units = time_units
            linear = True
            constraint_path = 't_phase'
        elif var_type == 'state':
            shape = phase.state_options[var]['shape']
            units = phase.state_options[var]['units']
            linear = False
            constraint_path = f'boundary_vals.{var}'
        elif var_type == 'control':
            shape = phase.control_options[var]['shape']
            units = phase.control_options[var]['units']
            linear = False
            constraint_path = f'control_values:{var}'
        elif var_type == 'parameter':
            shape = phase.parameter_options[var]['shape']
            units = phase.parameter_options[var]['units']
            linear = True
            constraint_path = f'parameter_vals:{var}'
        elif var_type in ('control_rate', 'control_rate2'):
            control_var = var[:-5] if var_type == 'control_rate' else var[:-6]
            shape = phase.control_options[control_var]['shape']
            control_units = phase.control_options[control_var]['units']
            d = 2 if var_type == 'control_rate2' else 1
            control_rate_units = get_rate_units(control_units, time_units, deriv=d)
            units = control_rate_units
            linear = False
            constraint_path = f'control_rates:{var}'
        else:
            # Failed to find variable, assume it is in the ODE. This requires introspection.
            constraint_path = f'boundary_vals.{var}'
            meta = get_source_metadata(ode_outputs, var, user_units=None, user_shape=None)
            shape = meta['shape']
            units = meta['units']
            linear = False

        return constraint_path, shape, units, linear

    def _get_num_timeseries_nodes(self):
        """
        Returns the number of nodes in the default timeseries for this transcription.

        Returns
        -------
        int
            The number of nodes in the default timeseries for this transcription.
        """
        return self.grid_data.num_nodes

    def _get_rate_source_path(self, state_name, phase):
        """
        Return the rate source location and indices for a given state name.

        Parameters
        ----------
        state_name : str
            Name of the state.
        phase : dymos.Phase
            Phase object containing the rate source.

        Returns
        -------
        str
            Path to the rate source.
        ndarray
            Array of source indices.
        """
        try:
            var = phase.state_options[state_name]['rate_source']
        except RuntimeError:
            raise ValueError(f"state '{state_name}' in phase '{phase.name}' was not given a "
                             "rate_source")

        # Note the rate source must be shape-compatible with the state
        var_type = phase.classify_var(var)

        # Determine the path to the variable
        if var_type == 't':
            rate_path = 't'
        elif var_type == 't_phase':
            rate_path = 't_phase'
        elif var_type == 'state':
            rate_path = f'states:{var}'
        elif var_type == 'control':
            rate_path = f'control_values:{var}'
        elif var_type == 'control_rate':
            control_name = var[:-5]
            rate_path = f'control_rates:{control_name}_rate'
        elif var_type == 'control_rate2':
            control_name = var[:-6]
            rate_path = f'control_rates:{control_name}_rate2'
        elif var_type == 'parameter':
            rate_path = f'parameter_vals:{var}'
        else:
            # Failed to find variable, assume it is in the ODE
            rate_path = f'ode_all.{var}'

        return rate_path

    def _get_timeseries_var_source(self, var, output_name, phase):
        """
        Return the source path and indices for a given variable to be connected to a timeseries.

        Parameters
        ----------
        var : str
            Name of the timeseries variable whose source is desired.
        output_name : str
            Name of the timeseries output whose source is desired.
        phase : dymos.Phase
            Phase object containing the variable, either as state, time, control, etc., or as an ODE output.

        Returns
        -------
        meta : dict
            Metadata pertaining to the variable at the given path. This dict contains 'src' (the path to the
            timeseries source), 'src_idxs' (an array of the
            source indices), 'units' (the units of the source variable), and 'shape' (the shape of the variable at
            a given node).
        """
        gd = self.grid_data
        var_type = phase.classify_var(var)
        time_units = phase.time_options['units']

        transcription = phase.options['transcription']
        ode = transcription._get_ode(phase)
        ode_outputs = get_promoted_vars(ode, 'output')

        # The default for node_idxs, applies to everything except states and parameters.
        node_idxs = gd.subset_node_indices['all']

        meta = {}

        # Determine the path to the variable
        if var_type == 't':
            path = 't'
            src_units = time_units
            src_shape = (1,)
        elif var_type == 't_phase':
            path = 't_phase'
            src_units = time_units
            src_shape = (1,)
        elif var_type == 'state':
            path = f'states:{var}'
            src_units = phase.state_options[var]['units']
            src_shape = phase.state_options[var]['shape']

            if gd.transcription == 'radau-ps':
                # Find the state_input indices which occur at segment endpoints, and repeat them twice
                state_input_idxs = gd.subset_node_indices['state_input']
                repeat_idxs = np.ones_like(state_input_idxs)
                if self.options['compressed']:
                    segment_end_idxs = gd.subset_node_indices['segment_ends'][1:-1]
                    # Repeat nodes that are on segment bounds (but not the first or last nodes in the phase)
                    nodes_to_repeat = list(set(state_input_idxs).intersection(set(segment_end_idxs)))
                    # Now find these nodes in the state input indices
                    idxs_of_ntr_in_state_inputs = np.where(np.isin(state_input_idxs, nodes_to_repeat))[0]
                    # All state input nodes are used once, but nodes_to_repeat are used twice
                    repeat_idxs[idxs_of_ntr_in_state_inputs] = 2
                # Now we have a way of mapping the state input indices to all nodes
                map_input_node_idxs_to_all = np.repeat(np.arange(gd.subset_num_nodes['state_input'],
                                                                 dtype=int), repeats=repeat_idxs)
                # Now select the subset of nodes we want to use.
                node_idxs = map_input_node_idxs_to_all[gd.subset_node_indices['all']]
        elif var_type == 'control':
            path = f'control_values:{var}'
            src_units = phase.control_options[var]['units']
            src_shape = phase.control_options[var]['shape']
        elif var_type == 'control_rate':
            control_name = var[:-5]
            path = f'control_rates:{control_name}_rate'
            control_name = var[:-5]
            src_units = get_rate_units(phase.control_options[control_name]['units'], time_units, deriv=1)
            src_shape = phase.control_options[control_name]['shape']
        elif var_type == 'control_rate2':
            control_name = var[:-6]
            path = f'control_rates:{control_name}_rate2'
            src_units = get_rate_units(phase.control_options[control_name]['units'], time_units, deriv=2)
            src_shape = phase.control_options[control_name]['shape']
        elif var_type == 'parameter':
            path = f'parameter_vals:{var}'
            # Timeseries are never a static_target
            node_idxs = np.zeros(gd.subset_num_nodes['all'], dtype=int)
            src_units = phase.parameter_options[var]['units']
            src_shape = phase.parameter_options[var]['shape']
        else:
            # Failed to find variable, assume it is in the ODE
            path = f'ode_all.{var}'
            meta = get_source_metadata(ode_outputs, src=var)
            src_shape = meta['shape']
            src_units = meta['units']
            src_tags = meta['tags']
            if 'dymos.static_output' in src_tags:
                raise RuntimeError(
                    f'ODE output {var} is tagged with "dymos.static_output" and cannot be a timeseries output.')

        src_idxs = om.slicer[node_idxs, ...]

        meta['src'] = path
        meta['src_idxs'] = src_idxs
        meta['units'] = src_units
        meta['shape'] = src_shape

        return meta

    def get_parameter_connections(self, name, phase):
        """
        Returns info about a parameter's target connections in the phase.

        Parameters
        ----------
        name : str
            Parameter name.
        phase : dymos.Phase
            The phase object to which this transcription instance applies.

        Returns
        -------
        list of (paths, indices)
            A list containing a tuple of target paths and corresponding src_indices to which the
            given design variable is to be connected.
        """
        connection_info = []
        if name in phase.parameter_options:
            options = phase.parameter_options[name]
            for tgt in options['targets']:
                if tgt in options['static_targets']:
                    src_idxs = np.squeeze(get_src_indices_by_row([0], options['shape']), axis=0)
                    endpoint_src_idxs = om.slicer[:, ...]
                else:
                    src_idxs_raw = np.zeros(self.grid_data.subset_num_nodes['all'], dtype=int)
                    src_idxs = get_src_indices_by_row(src_idxs_raw, options['shape'])
                    endpoint_src_idxs_raw = np.zeros(2, dtype=int)
                    endpoint_src_idxs = get_src_indices_by_row(endpoint_src_idxs_raw, options['shape'])
                    if options['shape'] == (1,):
                        src_idxs = src_idxs.ravel()
                        endpoint_src_idxs = endpoint_src_idxs.ravel()

                connection_info.append((f'ode_all.{tgt}', (src_idxs,)))
                connection_info.append((f'boundary_vals.{tgt}', (endpoint_src_idxs,)))

        return connection_info

    def _requires_continuity_constraints(self, phase):
        """
        Tests whether state and/or control and/or control rate continuity are required.

        Parameters
        ----------
        phase : dymos.Phase
            The phase to which this transcription applies.

        Returns
        -------
        any_state_continuity : bool
            True if any state continuity is required to be enforced.
        any_control_continuity : bool
            True if any control value continuity is required to be enforced.
        any_control_rate_continuity : bool
            True if any control rate continuity is required to be enforced.
        """
        num_seg = self.grid_data.num_segments
        compressed = self.grid_data.compressed

        any_state_continuity = num_seg > 1 and not compressed
        any_control_continuity = any([opts['continuity'] for opts in phase.control_options.values()])
        any_control_continuity = any_control_continuity and num_seg > 1
        any_rate_continuity = any([opts['rate_continuity'] or opts['rate2_continuity']
                                   for opts in phase.control_options.values()])
        any_rate_continuity = any_rate_continuity and num_seg > 1

        return any_state_continuity, any_control_continuity, any_rate_continuity

    def _phase_set_state_val(self, phase, name, vals, time_vals, interpolation_kind):
        """
        Method to interpolate the provided input and return the variables that need to be set
        along with their appropriate value.

        Parameters
        ----------
        phase : dymos.Phase
            The phase to which this transcription applies.
        name : str
            The name of the phase variable to be set.
        vals : ndarray or Sequence or float
            Array of control/state/parameter values.
        time_vals : ndarray or Sequence or None
            Array of integration variable values.
        interpolation_kind : str
            Specifies the kind of interpolation, as per the scipy.interpolate package.
            One of ('linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic'
            where 'zero', 'slinear', 'quadratic' and 'cubic' refer to a spline
            interpolation of zeroth, first, second or third order) or as an
            integer specifying the order of the spline interpolator to use.
            Default is 'linear'.

        Returns
        -------
        input_data : dict
            Dict containing the values that need to be set in the phase

        """
        input_data = {}
        if np.isscalar(vals):
            input_data[f'states:{name}'] = vals
            input_data[f'initial_states:{name}'] = vals
            input_data[f'final_states:{name}'] = vals
        else:
            interp_vals = phase.interp(name, vals, time_vals,
                                       nodes='state_input',
                                       kind=interpolation_kind)
            input_data[f'states:{name}'] = interp_vals
            input_data[f'initial_states:{name}'] = vals[0]
            input_data[f'final_states:{name}'] = vals[-1]

        return input_data



import openmdao.api as om
import dymos as dm
import unittest
import numpy as np

from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from dymos.examples.low_thrust_spiral import LowThrustODE

show_plots = False


def low_thrust_spiral_direct_collocation(grid_type='lgl'):

    optimizer = 'SNOPT'
    p = om.Problem(model=om.Group())
    p.driver = om.pyOptSparseDriver()
    p.driver.options['optimizer'] = optimizer
    p.driver.declare_coloring()
    if optimizer == 'SNOPT':
        p.driver.opt_settings['Major iterations limit'] = 1000
        p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
        p.driver.opt_settings['Major optimality tolerance'] = 5.0E-4
        p.driver.opt_settings['iSumm'] = 6

    t = dm.Birkhoff(num_nodes=101, grid_type=grid_type)

    traj = p.model.add_subsystem('traj', dm.Trajectory())

    phase = traj.add_phase('phase0', dm.Phase(ode_class=LowThrustODE, transcription=t))

    phase.set_time_options(fix_initial=True, fix_duration=False, units='s', duration_bounds=(100, 500))

    phase.add_state('r', fix_initial=True, fix_final=True, rate_source='vr', lower=0.5, upper=6.5,
                    defect_ref=1e2)
    phase.add_state('theta', fix_initial=True, fix_final=False, rate_source='theta_dot', units='rad', defect_ref=1e2)
    phase.add_state('vr', fix_initial=True, fix_final=True, rate_source='vr_dot', units='1/s', defect_ref=1e2)
    phase.add_state('vt', fix_initial=True, fix_final=True, rate_source='vt_dot', units='1/s', defect_ref=1e2)

    phase.add_control('alpha', units='rad')

    # Minimize the control effort
    phase.add_objective('time', loc='final')

    p.model.linear_solver = om.DirectSolver()

    p.setup(check=True, force_alloc_complex=True)

    phase.set_time_val(initial=0.0, duration=300.0)
    phase.set_state_val('r', [1.0, 6.0])
    phase.set_state_val('theta', [0.0, 0.0])
    phase.set_state_val('vr', [0.0, 0.0])
    phase.set_state_val('vt', [1.0, 1/np.sqrt(6)])
    phase.set_control_val('alpha', [0.0, 0.0])

    return p


@require_pyoptsparse(optimizer='SNOPT')
@use_tempdirs
class TestLowThrustSpiral(unittest.TestCase):

    @staticmethod
    def _assert_results(p, tol=0.05):
        t = p.get_val('traj.phase0.timeseries.time')
        assert_near_equal(t[-1], 228, tolerance=tol)
        return

    @unittest.skip('Long running test skipped on CI.')
    def test_low_thrust_spiral_lgl(self):
        p = low_thrust_spiral_direct_collocation(grid_type='lgl')
        dm.run_problem(p)
        theta = p.get_val('traj.phase0.timeseries.theta')
        r = p.get_val('traj.phase0.timeseries.r')

        if show_plots:
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots(subplot_kw={'projection': 'polar'})
            ax.plot(theta, r)
            ax.grid(True)
            plt.show()

        self._assert_results(p)

    @unittest.skip('Long running test skipped on CI.')
    def test_low_thrust_spiral_cgl(self):
        p = low_thrust_spiral_direct_collocation(grid_type='cgl')
        dm.run_problem(p)
        theta = p.get_val('traj.phase0.timeseries.theta')
        r = p.get_val('traj.phase0.timeseries.r')

        if show_plots:
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots(subplot_kw={'projection': 'polar'})
            ax.plot(theta, r)
            ax.grid(True)
            plt.show()

        self._assert_results(p)

    def test_check_partials(self):
        p = low_thrust_spiral_direct_collocation(grid_type='cgl')
        p.run_model()
        cpd = p.check_partials(method='cs', out_stream=None)
        assert_check_partials(cpd)


if __name__ == "__main__":
    unittest.main()



import numpy as np

_cgl_cache = {}
""" Cache for the LGL nodes and weights, keyed by n. """


def clenshaw_curtis(n):
    """
    Returns the Chebyshev-Gauss-Lobatto nodes and weights for a Jacobi Polynomial with n abscissae.

    The nodes are on the range [-1, 1].

    Parameters
    ----------
    n : int
        The number of LGL nodes requested.  The order of the polynomial is n-1.

    Returns
    -------
    x : numpy.array
        An array of the LGL nodes for a polynomial of the given order.

    w : numpy.array
        An array of the corresponding CGL weights at the nodes in x.
    """
    x = -np.cos(np.pi * np.arange(n) / (n-1))
    w = np.zeros(n)
    even = False if n % 2 else True

    N_2 = int(n/2) if even else int((n+1)/2)
    j_lim = N_2 if even else int((n-1)/2)

    for k in range(n):
        c = 1 if k == 0 else 2
        s = 0
        for j in range(1, N_2):
            b = 1 if j == j_lim else 2
            s += b * np.cos(2*j*k*np.pi/(n-1)) / (4*j**2 - 1)
        w[k] = c * (1 - s) / (n-1)

    if even:
        w[N_2:] = np.flip(w[:N_2])
    else:
        w[N_2:] = np.flip(w[:(N_2-1)])

    return x, w


def _cgl(n):
    """
    Returns the Chebyshev-Gauss-Lobatto nodes and weights for a Jacobi Polynomial with n abscissae.

    The nodes are on the range [-1, 1].

    Parameters
    ----------
    n : int
        The number of CGL nodes requested.  The order of the polynomial is n-1.

    Returns
    -------
    x : numpy.array
        An array of the CGL nodes for a polynomial of the given order.

    w : numpy.array
        An array of the corresponding CGL weights at the nodes in x.
    """
    x = -np.cos(np.pi * np.arange(n) / (n-1))
    w = np.pi / (n-1) * np.ones(n)
    w[0] *= 0.5
    w[-1] *= 0.5

    return x, w


def cgl(n):
    """
    Retrieve the cgl nodes and weights for n nodes.

    Results are cached to avoid repeated calculation of nodes and weights for a given n.

    Parameters
    ----------
    n : int
        Node number.

    Returns
    -------
    float
        Tuple with cgl nodes and weights.
    """
    if n not in _cgl_cache:
        _cgl_cache[n] = _cgl(n)
    return _cgl_cache[n]



import unittest

try:
    import matplotlib
    import matplotlib.pyplot as plt

    matplotlib.use('Agg')
    plt.style.use('ggplot')
except ImportError:
    matplotlib = None

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse


@use_tempdirs
class TestMinTimeClimbForDocs(unittest.TestCase):

    @require_pyoptsparse(optimizer='IPOPT')
    def test_min_time_climb_partial_coloring(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal

        import dymos as dm
        from dymos.examples.min_time_climb.doc.min_time_climb_ode_partial_coloring import MinTimeClimbODE

        #
        # Instantiate the problem and configure the optimization driver
        #
        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'IPOPT'
        p.driver.opt_settings['print_level'] = 0
        p.driver.opt_settings['tol'] = 1.0E-5
        p.driver.declare_coloring(tol=1.0E-12)

        #
        # Instantiate the trajectory and phase
        #
        traj = dm.Trajectory()

        phase = dm.Phase(ode_class=MinTimeClimbODE,
                         ode_init_kwargs={'fd': True, 'partial_coloring': True},
                         transcription=dm.GaussLobatto(num_segments=30))

        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj', traj)

        #
        # Set the options on the optimization variables
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(50, 400),
                               duration_ref=100.0)

        phase.add_state('r', fix_initial=True, lower=0, upper=1.0E6,
                        ref=1.0E3, defect_ref=1.0E3, units='m',
                        rate_source='flight_dynamics.r_dot')

        phase.add_state('h', fix_initial=True, lower=0, upper=20000.0,
                        ref=1.0E2, defect_ref=1.0E2, units='m',
                        rate_source='flight_dynamics.h_dot')

        phase.add_state('v', fix_initial=True, lower=10.0,
                        ref=1.0E2, defect_ref=1.0E2, units='m/s',
                        rate_source='flight_dynamics.v_dot')

        phase.add_state('gam', fix_initial=True, lower=-1.5, upper=1.5,
                        ref=1.0, defect_ref=1.0, units='rad',
                        rate_source='flight_dynamics.gam_dot')

        phase.add_state('m', fix_initial=True, lower=10.0, upper=1.0E5,
                        ref=1.0E3, defect_ref=1.0E3, units='kg',
                        rate_source='prop.m_dot')

        phase.add_control('alpha', units='deg', lower=-8.0, upper=8.0, scaler=1.0,
                          rate_continuity=True, rate_continuity_scaler=100.0,
                          rate2_continuity=False, targets=['alpha'])

        phase.add_parameter('S', val=49.2386, units='m**2', opt=False, targets=['S'])
        phase.add_parameter('Isp', val=1600.0, units='s', opt=False, targets=['Isp'])
        phase.add_parameter('throttle', val=1.0, opt=False, targets=['throttle'])

        #
        # Setup the boundary and path constraints
        #
        phase.add_boundary_constraint('h', loc='final', equals=20000, scaler=1.0E-3)
        phase.add_boundary_constraint('aero.mach', loc='final', equals=1.0)
        phase.add_boundary_constraint('gam', loc='final', equals=0.0)

        phase.add_path_constraint(name='h', lower=100.0, upper=20000, ref=20000)
        phase.add_path_constraint(name='aero.mach', lower=0.1, upper=1.8)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', ref=100.0)

        p.model.linear_solver = om.DirectSolver()

        #
        # Setup the problem and set the initial guess
        #
        p.setup()

        phase.set_time_val(initial=0.0, duration=350)
        phase.set_state_val('r', [0.0, 50000.0])
        phase.set_state_val('h', [100.0, 20000.0])
        phase.set_state_val('v', [135.964, 283.159])
        phase.set_state_val('gam', [0.0, 0.0])
        phase.set_state_val('m', [19030.468, 10000.])
        phase.set_control_val('alpha', [0.0, 0.0])

        #
        # Solve for the optimal trajectory
        #
        dm.run_problem(p)

        #
        # Test the results
        #
        assert_near_equal(p.get_val('traj.phase0.t_duration'), 321.0, tolerance=1.0E-1)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import os
import unittest

from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.mpi import MPI
from dymos.utils.misc import om_version


@use_tempdirs
class TestVanderpolForDocs(unittest.TestCase):
    def tearDown(self):
        for filename in ['total_coloring.pkl', 'SLSQP.out', 'SNOPT_print.out', 'SNOPT_summary.out']:
            if os.path.exists(filename):
                os.remove(filename)

    def test_vanderpol_for_docs_simulation(self):
        import dymos as dm
        from dymos.examples.vanderpol.vanderpol_dymos import vanderpol

        # Create the Dymos problem instance
        p = vanderpol(transcription='gauss-lobatto', num_segments=75)

        dm.run_problem(p, run_driver=False, simulate=True, make_plots=True)

    @unittest.skipIf(om_version()[0] < (3, 29, 0), 'Test requires OpenMDAO 3.29.0 or later')
    def test_vanderpol_for_docs_optimize(self):
        import dymos as dm
        from dymos.examples.vanderpol.vanderpol_dymos import vanderpol

        # Create the Dymos problem instance
        p = vanderpol(transcription='gauss-lobatto', num_segments=75,
                      transcription_order=3, compressed=True, optimizer='SLSQP')

        dm.run_problem(p, simulate=True, make_plots=True)

    @unittest.skipIf(om_version()[0] < (3, 29, 0), 'Test requires OpenMDAO 3.29.0 or later')
    def test_vanderpol_for_docs_optimize_refine(self):
        import dymos as dm
        from dymos.examples.vanderpol.vanderpol_dymos import vanderpol
        from openmdao.utils.assert_utils import assert_near_equal

        # Create the Dymos problem instance
        p = vanderpol(transcription='gauss-lobatto', num_segments=15,
                      transcription_order=3, compressed=True, optimizer='SLSQP')

        # Enable grid refinement and find optimal control solution to stop oscillation
        p.model.traj.phases.phase0.set_refine_options(refine=True)

        dm.run_problem(p, refine_iteration_limit=10, simulate=True, make_plots=True)

        assert_near_equal(p.get_val('traj.phase0.states:x0')[-1, ...], 0.0)
        assert_near_equal(p.get_val('traj.phase0.states:x1')[-1, ...], 0.0)
        assert_near_equal(p.get_val('traj.phase0.states:J')[-1, ...], 5.2808, tolerance=1.0E-3)
        assert_near_equal(p.get_val('traj.phase0.controls:u')[-1, ...], 0.0, tolerance=1.0E-3)


@unittest.skipUnless(MPI, "MPI is required.")
@use_tempdirs
class TestVanderpolDelayMPI(unittest.TestCase):
    N_PROCS = 2

    @unittest.skipIf(om_version()[0] < (3, 29, 0), 'Test requires OpenMDAO 3.29.0 or later')
    def test_vanderpol_delay_mpi(self):
        import openmdao.api as om
        import dymos as dm
        from dymos.examples.vanderpol.vanderpol_ode import VanderpolODE
        from openmdao.utils.assert_utils import assert_near_equal

        DELAY = 0.005

        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()

        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        # define a Trajectory object and add to model
        traj = dm.Trajectory()
        p.model.add_subsystem('traj', subsys=traj)

        t = dm.Radau(num_segments=30, order=3)

        # define a Phase as specified above and add to Phase
        phase = dm.Phase(ode_class=VanderpolODE, transcription=t,
                         ode_init_kwargs={'delay': DELAY, 'distrib': True})
        traj.add_phase(name='phase0', phase=phase)

        t_final = 15
        phase.set_time_options(fix_initial=True, fix_duration=True, duration_val=t_final, units='s')

        # set the State time options
        phase.add_state('x0', fix_initial=False, fix_final=False,
                        rate_source='x0dot',
                        units='V/s',
                        targets='x0')  # target required because x0 is an input
        phase.add_state('x1', fix_initial=False, fix_final=False,
                        rate_source='x1dot',
                        units='V',
                        targets='x1')  # target required because x1 is an input
        phase.add_state('J', fix_initial=False, fix_final=False,
                        rate_source='Jdot',
                        units=None)

        # define the control
        phase.add_control(name='u', units=None, lower=-0.75, upper=1.0, continuity=True,
                          rate_continuity=True,
                          targets='u')  # target required because u is an input

        # add constraints
        phase.add_boundary_constraint('x0', loc='initial', equals=1.0)
        phase.add_boundary_constraint('x1', loc='initial', equals=1.0)
        phase.add_boundary_constraint('J', loc='initial', equals=0.0)

        phase.add_boundary_constraint('x0', loc='final', equals=0.0)
        phase.add_boundary_constraint('x1', loc='final', equals=0.0)

        # define objective to minimize
        phase.add_objective('J', loc='final')

        # setup the problem
        p.setup(check=True)

        phase.set_time_val(0.0, t_final)

        # add a linearly interpolated initial guess for the state and control curves
        phase.set_state_val('x0', [1, 0])
        phase.set_state_val('x0', [1, 0])
        phase.set_state_val('x0', [0, 1])
        phase.set_control_val('u', -0.75)

        dm.run_problem(p, run_driver=True, simulate=False)

        assert_near_equal(p.get_val('traj.phase0.states:x0')[-1, ...], 0.0)
        assert_near_equal(p.get_val('traj.phase0.states:x1')[-1, ...], 0.0)
        assert_near_equal(p.get_val('traj.phase0.states:J')[-1, ...], 5.2808, tolerance=0.0025)
        assert_near_equal(p.get_val('traj.phase0.controls:u')[-1, ...], 0.0, tolerance=1.0E-3)



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from openmdao.utils.assert_utils import assert_near_equal
from dymos.utils.misc import om_version

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_vector_states_ode import BrachistochroneVectorStatesODE


@use_tempdirs
class TestAddBoundaryConstraint(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_simple_no_exception(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'

        p.driver.declare_coloring()

        transcription = dm.GaussLobatto(num_segments=3,
                                        order=3,
                                        compressed=True)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=transcription)
        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj0', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

        # can't fix final position if you're solving the segments
        phase.add_state('pos',
                        rate_source='pos_dot', units='m',
                        fix_initial=True)

        # test add_boundary_constraint with arrays:
        expected = np.array([10, 5])
        phase.add_boundary_constraint(name='pos', loc='final', equals=expected)

        phase.add_state('v',
                        rate_source='vdot', units='m/s',
                        fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9, opt=False)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)
        p.final_setup()

        pos0 = [0, 10]
        posf = [10, 5]

        phase.set_time_val(initial=0, duration=1.8016)
        phase.set_state_val('pos', (pos0, posf))
        phase.set_state_val('v', (0, 9.9))
        phase.set_control_val('theta', (5, 100))
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        assert_near_equal(p.get_val('traj0.phase0.timeseries.pos')[-1, ...],
                          [10, 5], tolerance=1.0E-5)

    def test_invalid_expression(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'

        p.driver.declare_coloring()

        transcription = dm.GaussLobatto(num_segments=3,
                                        order=3,
                                        compressed=True)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=transcription)
        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj0', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

        phase.add_state('pos',
                        rate_source='pos_dot', units='m',
                        fix_initial=True)

        phase.add_state('v',
                        rate_source='vdot', units='m/s',
                        fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665, opt=False)

        # test add_boundary_constraint with arrays:
        phase.add_boundary_constraint(name='pos**2', loc='final', equals=np.array([10, 5]))

        expected = "Unable to find the source 'pos**2' in the ODE."
        with self.assertRaises(ValueError) as e:
            p.setup()

        self.assertEqual(expected, str(e.exception))

    def test_duplicate_name(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'

        p.driver.declare_coloring()

        transcription = dm.GaussLobatto(num_segments=3,
                                        order=3,
                                        compressed=True)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=transcription)
        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj0', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

        # can't fix final position if you're solving the segments
        phase.add_state('pos',
                        rate_source='pos_dot', units='m',
                        fix_initial=True)

        phase.add_state('v',
                        rate_source='vdot', units='m/s',
                        fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665, opt=False)

        phase.add_boundary_constraint(name='pos', loc='final', equals=np.array([10, 5]))

        # test add_boundary_constraint with arrays:
        with self.assertRaises(ValueError) as e:
            phase.add_boundary_constraint(name='pos=v**2', loc='final', equals=np.array([10, 5]))

        expected = 'Cannot add new final boundary constraint for variable `pos` and indices None.' \
                   ' One already exists.'
        self.assertEqual(expected, str(e.exception))

    def test_duplicate_constraint(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.options['optimizer'] = 'SLSQP'

        p.driver.declare_coloring()

        transcription = dm.GaussLobatto(num_segments=3,
                                        order=3,
                                        compressed=True)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=transcription)
        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj0', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

        # can't fix final position if you're solving the segments
        phase.add_state('pos',
                        rate_source='pos_dot', units='m',
                        fix_initial=True)

        phase.add_state('v',
                        rate_source='vdot', units='m/s',
                        fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665, opt=False)

        phase.add_boundary_constraint(name='pos', loc='final', equals=np.array([10, 5]))

        # test add_boundary_constraint with arrays:
        with self.assertRaises(ValueError) as e:
            phase.add_boundary_constraint(name='pos', loc='final', equals=np.array([10, 5]))

        expected = 'Cannot add new final boundary constraint for variable `pos` and indices None. One already exists.'
        self.assertEqual(expected, str(e.exception))


if __name__ == '__main__':
    unittest.main()



from copy import deepcopy
import numpy as np
import openmdao.api as om

from dymos.transcriptions.explicit_shooting.vandermonde_control_interp_comp import VandermondeControlInterpComp
from dymos.transcriptions.explicit_shooting.cubic_spline_control_interp_comp import CubicSplineControlInterpComp

from .barycentric_control_interp_comp import BarycentricControlInterpComp
from .state_rate_collector_comp import StateRateCollectorComp
from .tau_comp import TauComp

from ...utils.introspection import configure_controls_introspection, \
    configure_time_introspection, configure_parameters_introspection, \
    configure_states_discovery, configure_states_introspection, _get_targets_metadata, \
    _get_common_metadata, get_promoted_vars
from ...utils.misc import get_rate_units, _unspecified, _none_or_unspecified
from ...utils.ode_utils import _make_ode_system


class ODEEvaluationGroup(om.Group):
    """
    A group whose purpose is to evaluate the ODE and output the computed state rates.

    Parameters
    ----------
    ode_class : class
        The class of the OpenMDAO system to be used to evaluate the ODE in this Group.
    input_grid_data : GridData
        The GridData used to define the controls used in this ODE.
    time_options : OptionsDictionary
        OptionsDictionary of time options.
    state_options : dict of {str: OptionsDictionary}
        For each state variable, a dictionary of its options, keyed by name.
    parameter_options : dict of {str: OptionsDictionary}
        For each parameter, a dictionary of its options, keyed by name.
    control_options : dict of {str: OptionsDictionary}
        For each control variable, a dictionary of its options, keyed by name.
    ode_init_kwargs : dict
        A dictionary of keyword arguments to be passed to the instantiation of the ODE.
    compute_derivs : bool
        If True, the derivatives need to be computed for propagation. In some cases,
        signficant setup time can be saved by skipping derivatives if not needed, such as during
        explicit simulation for verification.
    vec_size : int
        The number of points at which the ODE is simultaneously evaluated.
    control_interp : str
        The control interpolation technique to be used. Must be either 'vandermonde' or 'barycentric'.
    calc_exprs : dict
        A dictionary of ODE expressions.
    **kwargs : dict
        Additional keyword arguments passed to Group.
    """

    def __init__(self, ode_class, input_grid_data, time_options, state_options, parameter_options, control_options,
                 ode_init_kwargs=None, compute_derivs=True, vec_size=1,
                 control_interp='vandermonde', calc_exprs=None, **kwargs):
        super().__init__(**kwargs)

        # This component creates copies of the variable options from the phase.
        # It needs to perform its own introspection with respect to its ODE instance,
        # and this would override unspecified variables for parameter introspection
        # at the phase level.
        self._state_options = deepcopy(state_options)
        self._parameter_options = deepcopy(parameter_options)
        self._time_options = deepcopy(time_options)
        self._control_options = deepcopy(control_options)

        self._control_interpolants = {}
        self._ode_class = ode_class
        self._input_grid_data = input_grid_data
        self._compute_derivs = compute_derivs
        self._vec_size = vec_size
        self._ode_init_kwargs = {} if ode_init_kwargs is None else ode_init_kwargs
        self._calc_exprs = {} if calc_exprs is None else calc_exprs
        self._control_interp = control_interp

    def set_segment_index(self, seg_idx):
        """
        Set the segment_index option on those subsystems which require it.

        Parameters
        ----------
        seg_idx : int
            The index of the current segment.
        """
        self._get_subsystem('tau_comp').options['segment_index'] = seg_idx

        control_interp_comp = self._get_subsystem('control_interp')
        if control_interp_comp:
            control_interp_comp.set_segment_index(seg_idx)

    def setup(self):
        """
        Define the structure of the ODEEvaluationGroup.
        """
        igd = self._input_grid_data
        t_name = self._time_options['name']
        t_units = self._time_options['units']

        # All states, controls, and parameters need to exist
        # in the ODE evaluation group regardless of whether or not they have targets in the ODE.
        # This makes taking the derivatives more consistent without Exceptions.
        self._ivc = self.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])

        # Add a component to compute the current non-dimensional phase time.
        self.add_subsystem('tau_comp',
                           TauComp(grid_data=self._input_grid_data, vec_size=self._vec_size, time_units=t_units),
                           promotes_inputs=[('t', t_name), 't_initial', 't_duration'],
                           promotes_outputs=['stau', 'ptau', 'dstau_dt', ('t_phase', f'{t_name}_phase')])

        if self._control_options:
            c_options = self._control_options

            # Add control interpolant
            if self._control_interp == 'barycentric':
                self._control_comp = self.add_subsystem('control_interp',
                                                        BarycentricControlInterpComp(grid_data=igd,
                                                                                     control_options=c_options,
                                                                                     time_units=t_units,
                                                                                     compute_derivs=self._compute_derivs),
                                                        promotes_inputs=['ptau', 'stau', 't_duration', 'dstau_dt'])
            elif self._control_interp == 'vandermonde':
                self._control_comp = self.add_subsystem('control_interp',
                                                        VandermondeControlInterpComp(grid_data=igd,
                                                                                     control_options=c_options,
                                                                                     time_units=t_units,
                                                                                     compute_derivs=self._compute_derivs),
                                                        promotes_inputs=['ptau', 'stau', 't_duration', 'dstau_dt'])
            else:
                self._control_comp = self.add_subsystem('control_interp',
                                                        CubicSplineControlInterpComp(grid_data=igd,
                                                                                     control_options=c_options,
                                                                                     time_units=t_units,
                                                                                     compute_derivs=self._compute_derivs),
                                                        promotes_inputs=['ptau', 'stau', 'dstau_dt', 't_duration'])

        ode = _make_ode_system(ode_class=self._ode_class,
                               num_nodes=self._vec_size,
                               ode_init_kwargs=self._ode_init_kwargs,
                               calc_exprs=self._calc_exprs,
                               parameter_options=self._parameter_options)

        self.add_subsystem('ode', ode)

        self.add_subsystem('state_rate_collector',
                           StateRateCollectorComp(state_options=self._state_options,
                                                  time_units=self._time_options['units'],
                                                  vec_size=self._vec_size))

    def configure(self):
        """
        Perform I/O creation for this group's underlying members.

        In dymos, this system sits within a subproblem and therefore isn't in the standard
        configuration chain.  We need to perform all of the introspection of the ODE here.
        """
        ode = self._get_subsystem('ode')

        configure_time_introspection(self._time_options, ode)
        self._configure_time()

        configure_parameters_introspection(self._parameter_options, ode)
        self._configure_params()

        configure_controls_introspection(self._control_options, ode,
                                         time_units=self._time_options['units'])
        self._configure_controls()

        if self._control_options:
            self._get_subsystem('control_interp').configure_io()

        configure_states_discovery(self._state_options, ode)
        configure_states_introspection(self._state_options, self._time_options, self._control_options,
                                       self._parameter_options, ode)
        self._configure_states()

        self.state_rate_collector.configure_io()

    def _configure_time(self):
        vec_size = self._vec_size
        t_name = self._time_options['name']
        targets = self._time_options['targets']
        t_phase_targets = self._time_options['time_phase_targets']
        t_initial_targets = self._time_options['t_initial_targets']
        t_duration_targets = self._time_options['t_duration_targets']
        units = self._time_options['units']

        self._ivc.add_output(t_name, shape=(vec_size,), units=units)
        self._ivc.add_output('t_initial', shape=(1,), units=units)
        self._ivc.add_output('t_duration', shape=(1,), units=units)

        self.add_design_var(t_name)
        self.add_design_var('t_initial')
        self.add_design_var('t_duration')

        for tgts, var in [(targets, t_name), (t_phase_targets, f'{t_name}_phase'),
                          (t_initial_targets, 't_initial'), (t_duration_targets, 't_duration')]:
            for t in tgts:
                self.promotes('ode', inputs=[(t, var)])
            if tgts:
                self.set_input_defaults(name=var,
                                        val=np.ones((1,)),
                                        units=units)

    def _configure_states(self):
        vec_size = self._vec_size

        for name, options in self._state_options.items():
            shape = options['shape']
            units = options['units']
            targets = options['targets'] if options['targets'] is not None else []
            rate_path = self._get_rate_source_path(name)
            var_name = f'states:{name}'

            self._ivc.add_output(var_name, shape=(vec_size,) + shape, units=units)
            self.add_design_var(var_name)

            # Promote targets from the ODE
            for tgt in targets:
                self.promotes('ode', inputs=[(tgt, var_name)])
            if targets:
                self.set_input_defaults(name=var_name,
                                        val=np.ones(shape),
                                        units=options['units'])

            self.connect(rate_path, f'state_rate_collector.state_rates_in:{name}_rate')

            self.add_constraint(f'state_rate_collector.state_rates:{name}_rate')

    def _configure_params(self):
        ode_inputs = get_promoted_vars(self.ode, iotypes='input', metadata_keys=['shape', 'units', 'val', 'tags'])

        for name, options in self._parameter_options.items():
            var_name = f'parameters:{name}'

            targets = _get_targets_metadata(ode_inputs, name=name, user_targets=options['targets'])

            if options['units'] is _unspecified:
                units = _get_common_metadata(targets, 'units')
            else:
                units = options['units']

            if options['shape'] in _none_or_unspecified:
                shape = _get_common_metadata(targets, 'shape')
            else:
                shape = options['shape']

            self._ivc.add_output(var_name, shape=shape, units=units)
            self.add_design_var(var_name)

            # Promote targets from the ODE
            for tgt in targets:
                if tgt in options['static_targets']:
                    shape = None
                self.promotes('ode', inputs=[(tgt, var_name)],
                              src_shape=shape)
            if targets:
                self.set_input_defaults(name=var_name,
                                        val=1.0,
                                        units=options['units'])

    def _configure_controls(self):
        configure_controls_introspection(self._control_options, self.ode)
        time_units = self._time_options['units']

        if self._control_options:
            igd = self._input_grid_data

            if igd is None:
                raise ValueError('ODEEvaluationGroup was provided with control options but '
                                 'a GridData object was not provided.')

            for name, options in self._control_options.items():
                if options['control_type'] == 'polynomial':
                    num_control_input_nodes = options['order'] + 1
                else:
                    num_control_input_nodes = igd.subset_num_nodes['control_input']

                shape = options['shape']
                units = options['units']
                rate_units = get_rate_units(units, time_units, deriv=1)
                rate2_units = get_rate_units(units, time_units, deriv=2)
                targets = options['targets']
                rate_targets = options['rate_targets']
                rate2_targets = options['rate2_targets']
                uhat_name = f'controls:{name}'
                u_name = f'control_values:{name}'
                u_rate_name = f'control_rates:{name}_rate'
                u_rate2_name = f'control_rates:{name}_rate2'

                self._ivc.add_output(uhat_name, shape=(num_control_input_nodes,) + shape, units=units)
                self.add_design_var(uhat_name)
                self.add_constraint(u_name)
                self.add_constraint(u_rate_name)
                self.add_constraint(u_rate2_name)

                self.promotes('control_interp', inputs=[uhat_name],
                              outputs=[u_name, u_rate_name, u_rate2_name])

                # Promote targets from the ODE
                for tgt in targets:
                    self.promotes('ode', inputs=[(tgt, u_name)])
                if targets:
                    self.set_input_defaults(name=u_name,
                                            val=np.ones(shape),
                                            units=options['units'])

                # Promote rate targets from the ODE
                for tgt in rate_targets:
                    self.promotes('ode', inputs=[(tgt, u_rate_name)])
                if rate_targets:
                    self.set_input_defaults(name=u_rate_name,
                                            val=np.ones(shape),
                                            units=rate_units)

                # Promote rate2 targets from the ODE
                for tgt in rate2_targets:
                    self.promotes('ode', inputs=[(tgt, u_rate2_name)])
                if rate2_targets:
                    self.set_input_defaults(name=u_rate2_name,
                                            val=np.ones(shape),
                                            units=rate2_units)

    def _get_rate_source_path(self, state_var):
        """
        Get path of the rate source variable so that we can connect it to the
        outputs when we're done.

        Parameters
        ----------
        state_var : str
            The name of the state variable whose path is desired.

        Returns
        -------
        path : str
            The path to the rate source of the state variable.
        io : str
            A string indicating whether the variable in the path is an 'input'
            or an 'output'.
        """
        var = self._state_options[state_var]['rate_source']
        t_name = self._time_options['name']

        if var == t_name:
            rate_path = t_name
        elif var == f'{t_name}_phase':
            rate_path = f'{t_name}_phase'
        elif self._state_options is not None and var in self._state_options:
            rate_path = f'states:{var}'
        elif self._control_options is not None and var in self._control_options:
            rate_path = f'control_values:{var}'
        elif self._parameter_options is not None and var in self._parameter_options:
            rate_path = f'parameters:{var}'
        elif var.endswith('_rate') and self._control_options is not None and \
                var[:-5] in self._control_options:
            rate_path = f'control_rates:{var}'
        elif var.endswith('_rate2') and self._control_options is not None and \
                var[:-6] in self._control_options:
            rate_path = f'control_rates:{var}'
        else:
            rate_path = f'ode.{var}'
        return rate_path



import unittest

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs

try:
    from openmdao.drivers.pyoptsparse_driver import pyOptSparseDriver
except Exception:
    pyOptSparseDriver = None

import dymos as dm


class ODEComp(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']
        # z is the state vector, a nn x 2 x 2 in the form of [[x, y], [vx, vy]]
        self.add_input('param', shape=(3,), units=None)
        self.add_input('z', shape=(nn, 2, 2), units=None)
        self.add_output('zdot', shape=(nn, 2, 2), units=None)

        self.declare_partials(of='zdot', wrt='z', method='cs')
        self.declare_coloring(wrt=['z'], method='cs', num_full_jacs=5, tol=1.0E-12)

    def compute(self, inputs, outputs):
        outputs['zdot'][:, 0, 0] = inputs['z'][:, 1, 0]
        outputs['zdot'][:, 0, 1] = inputs['z'][:, 1, 1]
        outputs['zdot'][:, 1, 0] = 0.0
        outputs['zdot'][:, 1, 1] = -9.81


def add_parameter_test(testShape=None):
    p = om.Problem()

    tx = dm.Radau(num_segments=10, order=3, solve_segments=False)

    traj = dm.Trajectory()
    phase = dm.Phase(ode_class=ODEComp, transcription=tx)

    if testShape is None:
        phase.add_parameter('param', static_target=True)
    else:
        phase.add_parameter('param', shape=testShape, static_target=True)

    traj.add_phase('phase', phase)

    p.model.add_subsystem('traj', traj)

    phase.set_time_options(fix_initial=True, duration_bounds=(1, 5), units=None)
    phase.add_state('z', rate_source='zdot', fix_initial=True, units=None)

    phase.add_boundary_constraint('z', loc='final', lower=0, upper=0, indices=[1])
    phase.add_objective('time', loc='final')

    p.driver = om.ScipyOptimizeDriver()
    p.driver.declare_coloring(tol=1.0E-12)

    p.setup()

    p['traj.phase.parameters:param'] = [2.5, 3.5, 4.5]

    p.set_val('traj.phase.t_initial', 0)
    p.set_val('traj.phase.t_duration', 5)
    p.set_val('traj.phase.states:z', phase.interp('z', [[[0, 0], [10, 10]], [[10, 0], [10, -10]]]))

    p.run_driver()


@use_tempdirs
class TestParameterTypes(unittest.TestCase):
    def test_tuple(self):
        add_parameter_test((3, ))

    def test_list(self):
        add_parameter_test([3, ])

    def test_scaler(self):
        add_parameter_test(3)

    def test_nothing(self):
        add_parameter_test()



import numpy as np

import openmdao.api as om


class FuelBurnRateComp(om.ExplicitComponent):
    """ Computes the fuel burn rate (rate of change of fuel weight) based on SFC and thrust. """
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input(name='thrust', shape=(nn,), desc='current thrust', units='N')
        self.add_input(name='tsfc', shape=(nn,), desc='specific fuel consumption', units='1/s')

        # Outputs
        self.add_output(name='dXdt:mass_fuel', shape=(nn,),
                        desc='rate of aircraft mass change - negative when fuel is being depleted',
                        units='kg/s')

        # Partials
        ar = np.arange(nn)
        self.declare_partials('dXdt:mass_fuel', 'thrust', rows=ar, cols=ar)
        self.declare_partials('dXdt:mass_fuel', 'tsfc', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        outputs['dXdt:mass_fuel'] = -inputs['tsfc'] * inputs['thrust'] / 9.80665

    def compute_partials(self, inputs, partials):
        partials['dXdt:mass_fuel', 'thrust'] = -inputs['tsfc'] / 9.80665
        partials['dXdt:mass_fuel', 'tsfc'] = -inputs['thrust'] / 9.80665
